/*-
 * %%----------------------------------------------------------------------------------------------
 * Archive ouverte UNIGE - AoU Access - PublicationDownloadService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2022 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.aou.service;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;

import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.stereotype.Service;

import ch.unige.solidify.config.SolidifyEventPublisher;
import ch.unige.solidify.exception.SolidifyFileDeleteException;
import ch.unige.solidify.specification.SolidifySpecification;

import ch.unige.aou.business.AouResourceService;
import ch.unige.aou.controller.AccessController;
import ch.unige.aou.message.PublicationDownloadMessage;
import ch.unige.aou.model.access.PublicationDownload;
import ch.unige.aou.repository.PublicationDownloadRepository;
import ch.unige.aou.specification.PublicationDownloadSpecification;

@Service
@ConditionalOnBean(AccessController.class)
public class PublicationDownloadService extends AouResourceService<PublicationDownload> {

  @Override
  public SolidifySpecification<PublicationDownload> getSpecification(PublicationDownload resource) {
    return new PublicationDownloadSpecification(resource);
  }

  public Optional<PublicationDownload> findByFinalData(URI finalData) {
    return ((PublicationDownloadRepository) this.itemRepository).findByFinalData(finalData);
  }

  public List<PublicationDownload> findByPublicationId(String resId) {
    return ((PublicationDownloadRepository) this.itemRepository).findByPublicationId(resId);
  }

  public PublicationDownload createAndSubmitPublicationDownload(String resId) {
    PublicationDownload publicationDownload = this.createPublicationDownload(resId);
    if (publicationDownload.getStatus() != PublicationDownload.DownloadStatus.READY) {
      SolidifyEventPublisher.getPublisher().publishEvent(new PublicationDownloadMessage(publicationDownload.getResId()));
    }
    return publicationDownload;
  }

  public PublicationDownload createPublicationDownload(String resId) {
    PublicationDownload publicationDownload = new PublicationDownload();
    publicationDownload.setResId(resId);
    publicationDownload.setStatus(PublicationDownload.DownloadStatus.SUBMITTED);

    return this.save(publicationDownload);
  }

  @Override
  public void delete(String resId) {
    PublicationDownload publicationDownload = this.findOne(resId);
    super.delete(resId);

    if (publicationDownload.getFinalData() != null) {
      try {
        Path parent = Path.of(publicationDownload.getFinalData()).getParent();
        Files.walk(parent)
                .sorted(Comparator.reverseOrder())
                .map(Path::toFile)
                .forEach(File::delete);
      } catch (IOException e) {
        throw new SolidifyFileDeleteException(e.getMessage());
      }
    }
  }

}
