/*-
 * %%----------------------------------------------------------------------------------------------
 * Archive ouverte UNIGE - AoU Admin - SystemPropertyController.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2022 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.aou.controller.admin;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.linkTo;
import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.methodOn;

import java.util.List;
import java.util.Map;

import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.hateoas.RepresentationModel;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import ch.unige.solidify.config.SolidifyProperties;
import ch.unige.solidify.controller.ControllerWithHateoasHome;
import ch.unige.solidify.model.index.FacetProperties;

import ch.unige.aou.config.AouProperties;
import ch.unige.aou.controller.AdminController;
import ch.unige.aou.model.settings.StaticPage;
import ch.unige.aou.rest.UrlPath;
import ch.unige.aou.service.rest.trusted.TrustedIndexFieldAliasRemoteResourceService;
import ch.unige.aou.service.settings.StaticPageService;

@RestController
@ConditionalOnBean(AdminController.class)
@RequestMapping(UrlPath.ADMIN_SYSTEM_PROPERTIES)
public class SystemPropertyController implements ControllerWithHateoasHome {

  private final SystemProperties systemProperties;
  private final TrustedIndexFieldAliasRemoteResourceService indexFieldAliasRemoteService;
  private final StaticPageService staticPageService;
  private final String indexName;

  private static class SystemProperties extends RepresentationModel<SystemProperties> {
    private final String defaultChecksum;
    private final String defaultLicense;
    private final String fileSizeLimit;
    private final int limitMaxContributors;
    private final String orcidClientId;
    private final Map<String, String> defaultContributorRoles;
    private List<FacetProperties> searchFacets;
    private List<StaticPage> staticPages;

    public SystemProperties(AouProperties aouProperties, SolidifyProperties solidifyProperties) {
      this.defaultChecksum = aouProperties.getParameters().getDefaultChecksum();
      this.defaultLicense = aouProperties.getParameters().getDefaultLicense();
      this.fileSizeLimit = String.valueOf(aouProperties.getParameters().getFileSizeLimit().toBytes());
      this.limitMaxContributors = aouProperties.getMetadata().getImports().getLimitMaxContributors();
      this.orcidClientId = solidifyProperties.getOrcid().getClientId();
      this.defaultContributorRoles = aouProperties.getMetadata().getDefault().getContributorRolesByDepositSubtype();
    }

    public String getDefaultChecksum() {
      return this.defaultChecksum;
    }

    public String getDefaultLicense() {
      return this.defaultLicense;
    }

    public String getFileSizeLimit() {
      return this.fileSizeLimit;
    }

    public int getLimitMaxContributors() {
      return this.limitMaxContributors;
    }

    public String getOrcidClientId() {
      return this.orcidClientId;
    }

    public Map<String, String> getDefaultContributorRoles() {
      return this.defaultContributorRoles;
    }

    public List<FacetProperties> getSearchFacets() {
      return this.searchFacets;
    }

    public void setSearchFacets(List<FacetProperties> searchFacets) {
      this.searchFacets = searchFacets;
    }

    public List<StaticPage> getStaticPages() {
      return this.staticPages;
    }

    public void setStaticPages(List<StaticPage> staticPages) {
      this.staticPages = staticPages;
    }
  }

  public SystemPropertyController(AouProperties aouProperties, SolidifyProperties solidifyProperties,
          TrustedIndexFieldAliasRemoteResourceService indexFieldAliasRemoteService, StaticPageService staticPageService) {
    this.indexFieldAliasRemoteService = indexFieldAliasRemoteService;
    this.staticPageService = staticPageService;
    this.systemProperties = new SystemProperties(aouProperties, solidifyProperties);
    this.systemProperties.add(linkTo(methodOn(this.getClass()).home()).withSelfRel());
    this.indexName = aouProperties.getIndexing().getPublishedPublicationsIndexName();
  }

  @GetMapping
  public HttpEntity<SystemProperties> systemProperties() {
    this.systemProperties.setSearchFacets(this.indexFieldAliasRemoteService.getFacetProperties(this.indexName));
    this.systemProperties.setStaticPages(this.staticPageService.getStaticPages());
    this.systemProperties.removeLinks();
    this.systemProperties.add(linkTo(this.getClass()).withSelfRel());
    return new ResponseEntity<>(this.systemProperties, HttpStatus.OK);
  }
}
